@@ -56,7 +56,7 @@ struct timeval seqno_time = {0, 0};
 static int
 known_ae(int ae)
 {
-    return ae <= 3;
+    return ae <= AE_IPV6_LOCAL || ae == AE_V4VIAV6;
 }
 
 /* Parse a network prefix, encoded in the somewhat baroque compressed
@@ -83,10 +83,11 @@ network_prefix(int ae, int plen, unsigned int omitted,
     memset(prefix, 0, 16);
 
     switch(ae) {
-    case 0:
+    case AE_WILDCARD:
         ret = 0;
         break;
-    case 1:
+    case AE_IPV4:
+    case AE_V4VIAV6:
         if(omitted > 4 || pb > 4 || (pb > omitted && len < pb - omitted))
             return -1;
         memcpy(prefix, v4prefix, 12);
@@ -97,7 +98,7 @@ network_prefix(int ae, int plen, unsigned int omitted,
         if(pb > omitted) memcpy(prefix + 12 + omitted, p, pb - omitted);
         ret = pb - omitted;
         break;
-    case 2:
+    case AE_IPV6:
         if(omitted > 16 || (pb > omitted && len < pb - omitted)) return -1;
         if(omitted) {
             if(dp == NULL || v4mapped(dp)) return -1;
@@ -106,7 +107,7 @@ network_prefix(int ae, int plen, unsigned int omitted,
         if(pb > omitted) memcpy(prefix + omitted, p, pb - omitted);
         ret = pb - omitted;
         break;
-    case 3:
+    case AE_IPV6_LOCAL:
         if(pb > 8 && len < pb - 8) return -1;
         prefix[0] = 0xfe;
         prefix[1] = 0x80;
@@ -117,7 +118,8 @@ network_prefix(int ae, int plen, unsigned int omitted,
         return -1;
     }
 
-    normalize_prefix(p_r, prefix, plen < 0 ? 128 : ae == 1 ? plen + 96 : plen);
+    normalize_prefix(p_r, prefix,
+                     plen < 0 ? 128 : ae_is_v4(ae) ? plen + 96 : plen);
     return ret;
 }
 
@@ -175,7 +177,7 @@ parse_update_subtlv(struct interface *ifp, int metric, int ae,
                                 len - 1, src_prefix);
             if(rc < 0)
                 goto fail;
-            if(ae == 1)
+            if(ae_is_v4(ae))
                 *src_plen = a[i + 2] + 96;
             else
                 *src_plen = a[i + 2];
@@ -341,7 +343,7 @@ parse_request_subtlv(int ae, const unsigned char *a, int alen,
                                 len - 1, src_prefix);
             if(rc < 0)
                 goto fail;
-            if(ae == 1)
+            if(ae_is_v4(ae))
                 *src_plen = a[i + 2] + 96;
             else
                 *src_plen = a[i + 2];
@@ -392,7 +394,7 @@ parse_seqno_request_subtlv(int ae, const unsigned char *a, int alen,
                                 len - 1, src_prefix);
             if(rc < 0)
                 goto fail;
-            if(ae == 1)
+            if(ae_is_v4(ae))
                 (*src_plen) += 96;
         } else {
             debugf("Received unknown%s Route Request sub-TLV %d.\n",
@@ -591,9 +593,10 @@ parse_packet(const unsigned char *from, struct interface *ifp,
     int bodylen;
     struct neighbour *neigh = NULL;
     int have_router_id = 0, have_v4_prefix = 0, have_v6_prefix = 0,
+        have_v4viav6_prefix = 0,
         have_v4_nh = 0, have_v6_nh = 0;
     unsigned char router_id[8], v4_prefix[16], v6_prefix[16],
-        v4_nh[16], v6_nh[16];
+        v4viav6_prefix[16], v4_nh[16], v6_nh[16];
     int have_hello_rtt = 0;
     /* Content of the RTT sub-TLV on IHU messages. */
     unsigned int hello_send_us = 0, hello_rtt_receive_time = 0;
@@ -745,7 +748,8 @@ parse_packet(const unsigned char *from, struct interface *ifp,
                    txcost, interval,
                    format_address(from), ifp->name,
                    format_address(address));
-            if(message[2] == 0 || interface_ll_address(ifp, address)) {
+            if(message[2] == AE_WILDCARD ||
+                    interface_ll_address(ifp, address)) {
                 int changed;
                 rc = parse_ihu_subtlv(message + 8 + rc, len - 6 - rc,
                                       &hello_send_us, &hello_rtt_receive_time,
@@ -800,12 +804,21 @@ parse_packet(const unsigned char *from, struct interface *ifp,
             debugf("Received nh %s (%d) from %s on %s.\n",
                    format_address(nh), message[2],
                    format_address(from), ifp->name);
-            if(message[2] == 1) {
+            switch(message[2]) {
+            case AE_IPV4:
                 memcpy(v4_nh, nh, 16);
                 have_v4_nh = 1;
-            } else {
+                break;
+            case AE_IPV6:
+            case AE_IPV6_LOCAL:
                 memcpy(v6_nh, nh, 16);
                 have_v6_nh = 1;
+                break;
+            case AE_V4VIAV6:
+                goto done;
+            default:
+                /* We should have guards against this before this point. */
+                goto fail;
             }
             rc = parse_other_subtlv(message + 4 + rc, len - 2 - rc);
             if(rc < 0)
@@ -819,7 +832,7 @@ parse_packet(const unsigned char *from, struct interface *ifp,
             int rc, parsed_len, is_ss;
             if(len < 10) {
                 if(len < 2 || message[3] & 0x80)
-                    have_v4_prefix = have_v6_prefix = 0;
+                    have_v4_prefix = have_v6_prefix = have_v4viav6_prefix = 0;
                 goto fail;
             }
             if(!known_ae(message[2])) {
@@ -827,18 +840,31 @@ parse_packet(const unsigned char *from, struct interface *ifp,
                        message[2]);
                 goto done;
             }
+            if(message[2] == AE_V4VIAV6 && !has_v4viav6) {
+                /* We can safely ignore the prefix update that might come
+                   alongside with this TLV, since we ignore every v4-via-v6
+                   TLVs */
+                debugf("Ignoring v4-via-v6 route (unsupported).\n");
+                goto done;
+            }
             DO_NTOHS(interval, message + 6);
             DO_NTOHS(seqno, message + 8);
             DO_NTOHS(metric, message + 10);
             if(message[5] == 0 ||
-               (message[2] == 1 ? have_v4_prefix : have_v6_prefix))
+               (message[2] == AE_IPV4 ? have_v4_prefix :
+                message[2] == AE_IPV6 ? have_v6_prefix :
+                message[2] == AE_V4VIAV6 ? have_v4viav6_prefix :
+                0))
                 rc = network_prefix(message[2], message[4], message[5],
                                     message + 12,
-                                    message[2] == 1 ? v4_prefix : v6_prefix,
+                                    (message[2] == AE_IPV4 ? v4_prefix :
+                                     message[2] == AE_IPV6 ? v6_prefix :
+                                     message[2] == AE_V4VIAV6 ?  v4viav6_prefix :
+                                     NULL),
                                     len - 10, prefix);
             else
                 rc = -1;
-            if(message[2] == 1) {
+            if(ae_is_v4(message[2])) {
                 v4tov6(src_prefix, zeroes);
                 src_plen = 96;
             } else {
@@ -847,24 +873,36 @@ parse_packet(const unsigned char *from, struct interface *ifp,
             }
             if(rc < 0) {
                 if(message[3] & 0x80)
-                    have_v4_prefix = have_v6_prefix = 0;
+                    have_v4_prefix = have_v6_prefix = have_v4viav6_prefix = 0;
                 goto fail;
             }
             parsed_len = 10 + rc;
 
-            plen = message[4] + (message[2] == 1 ? 96 : 0);
+            plen = message[4] + (ae_is_v4(message[2]) ? 96 : 0);
 
             if(message[3] & 0x80) {
-                if(message[2] == 1) {
+                switch(message[2]) {
+                case AE_IPV4:
                     memcpy(v4_prefix, prefix, 16);
                     have_v4_prefix = 1;
-                } else {
+                    break;
+                case AE_IPV6:
                     memcpy(v6_prefix, prefix, 16);
                     have_v6_prefix = 1;
+                    break;
+                case AE_V4VIAV6:
+                    memcpy(v4viav6_prefix, prefix, 16);
+                    have_v4viav6_prefix = 1;
+                    break;
+                default:
+                    debugf("Received default prefix update with invalid "
+                           "AE %d.\n",
+                           message[2]);
+                    break;
                 }
             }
             if(message[3] & 0x40) {
-                if(message[2] == 1) {
+                if(ae_is_v4(message[2])) {
                     memset(router_id, 0, 4);
                     memcpy(router_id + 4, prefix + 12, 4);
                 } else {
@@ -872,7 +910,9 @@ parse_packet(const unsigned char *from, struct interface *ifp,
                 }
                 have_router_id = 1;
             }
-            if(metric < INFINITY && !have_router_id && message[2] != 0) {
+            if(metric < INFINITY &&
+               !have_router_id &&
+               message[2] != AE_WILDCARD) {
                 fprintf(stderr, "Received prefix with no router id.\n");
                 goto fail;
             }
@@ -881,7 +921,7 @@ parse_packet(const unsigned char *from, struct interface *ifp,
                    (message[3] & 0x40) ? "/id" : "",
                    format_prefix(prefix, plen),
                    format_address(from), ifp->name);
-            if(message[2] == 1) {
+            if(message[2] == AE_IPV4) {
                 if(have_v4_nh) {
                     nh = v4_nh;
                 } else {
@@ -902,7 +942,7 @@ parse_packet(const unsigned char *from, struct interface *ifp,
             if(rc < 0)
                 goto done;
 
-            if(message[2] == 0) {
+            if(message[2] == AE_WILDCARD) {
                 if(metric < 0xFFFF) {
                     fprintf(stderr,
                             "Received wildcard update with finite metric.\n");
@@ -926,11 +966,6 @@ parse_packet(const unsigned char *from, struct interface *ifp,
                    is_ss ? format_prefix(src_prefix, src_plen) : "",
                    format_address(from), ifp->name);
 
-            if(message[2] == 1) {
-                if(!ifp->ipv4)
-                    goto done;
-            }
-
             update_route(have_router_id ? router_id : NULL,
                          prefix, plen, src_prefix, src_plen, seqno,
                          metric, interval, neigh, nh,
@@ -947,8 +982,8 @@ parse_packet(const unsigned char *from, struct interface *ifp,
             rc = network_prefix(message[2], message[3], 0,
                                 message + 4, NULL, len - 2, prefix);
             if(rc < 0) goto fail;
-            plen = message[3] + (message[2] == 1 ? 96 : 0);
-            if(message[2] == 1) {
+            plen = message[3] + (ae_is_v4(message[2]) ? 96 : 0);
+            if(ae_is_v4(message[2])) {
                 v4tov6(src_prefix, zeroes);
                 src_plen = 96;
             } else {
@@ -960,7 +995,7 @@ parse_packet(const unsigned char *from, struct interface *ifp,
             if(rc < 0)
                 goto done;
             is_ss = !is_default(src_prefix, src_plen);
-            if(message[2] == 0) {
+            if(message[2] == AE_WILDCARD) {
                 if(is_ss) {
                     /* Wildcard requests don't carry a source prefix. */
                     fprintf(stderr,
@@ -982,7 +1017,8 @@ parse_packet(const unsigned char *from, struct interface *ifp,
                 }
             } else {
                 debugf("Received request for dst %s%s%s from %s on %s.\n",
-                       message[2] == 0 ? "" : format_prefix(prefix, plen),
+                       message[2] == AE_WILDCARD ?
+                                     "" : format_prefix(prefix, plen),
                        is_ss ? " src " : "",
                        is_ss ? format_prefix(src_prefix, src_plen) : "",
                        format_address(from), ifp->name);
@@ -1002,7 +1038,7 @@ parse_packet(const unsigned char *from, struct interface *ifp,
             rc = network_prefix(message[2], message[3], 0,
                                 message + 16, NULL, len - 14, prefix);
             if(rc < 0) goto fail;
-            if(message[2] == 1) {
+            if(ae_is_v4(message[2])) {
                 v4tov6(src_prefix, zeroes);
                 src_plen = 96;
             } else {
@@ -1015,7 +1051,7 @@ parse_packet(const unsigned char *from, struct interface *ifp,
             if(rc < 0)
                 goto done;
             is_ss = !is_default(src_prefix, src_plen);
-            plen = message[3] + (message[2] == 1 ? 96 : 0);
+            plen = message[3] + (ae_is_v4(message[2]) ? 96 : 0);
             debugf("Received request (%d) for dst %s%s%s from %s on "
                    "%s (%s, %d).\n",
                    message[6],
@@ -1400,7 +1436,7 @@ really_buffer_update(struct buffered *buf, struct interface *ifp,
                      unsigned char *channels, int channels_len)
 {
     int add_metric, v4, real_plen, real_src_plen;
-    int omit, spb, channels_size, len;
+    int ae, omit, spb, channels_size, len;
     const unsigned char *real_prefix, *real_src_prefix;
     unsigned short flags = 0;
     int is_ss = !is_default(src_prefix, src_plen);
@@ -1424,24 +1460,30 @@ really_buffer_update(struct buffered *buf, struct interface *ifp,
     v4 = plen >= 96 && v4mapped(prefix);
 
     if(v4) {
-        if(!ifp->ipv4)
-            return;
-        omit = 0;
-        if(!buf->have_nh ||
-           memcmp(buf->nh, ifp->ipv4, 4) != 0) {
-            start_message(buf, ifp, MESSAGE_NH, 6);
-            accumulate_byte(buf, 1);
-            accumulate_byte(buf, 0);
-            accumulate_bytes(buf, ifp->ipv4, 4);
-            end_message(buf, MESSAGE_NH, 6);
-            memcpy(&buf->nh, ifp->ipv4, 4);
-            buf->have_nh = 1;
+        if(!ifp->ipv4) {
+            if((ifp->flags & IF_V4VIAV6) == 0)
+                return;
+            ae = AE_V4VIAV6;
+        } else {
+            ae = AE_IPV4;
+            if(!buf->have_nh ||
+               memcmp(buf->nh, ifp->ipv4, 4) != 0) {
+                start_message(buf, ifp, MESSAGE_NH, 6);
+                accumulate_byte(buf, AE_IPV4);
+                accumulate_byte(buf, 0);
+                accumulate_bytes(buf, ifp->ipv4, 4);
+                end_message(buf, MESSAGE_NH, 6);
+                memcpy(&buf->nh, ifp->ipv4, 4);
+                buf->have_nh = 1;
+            }
         }
+        omit = 0;
         real_prefix = prefix + 12;
         real_plen = plen - 96;
         real_src_prefix = src_prefix + 12;
         real_src_plen = src_plen - 96;
     } else {
+        ae = AE_IPV6;
         omit = 0;
         if(buf->have_prefix) {
             while(omit < plen / 8 &&
@@ -1477,7 +1519,7 @@ really_buffer_update(struct buffered *buf, struct interface *ifp,
         len += 3 + spb;
 
     start_message(buf, ifp, MESSAGE_UPDATE, len);
-    accumulate_byte(buf, v4 ? 1 : 2);
+    accumulate_byte(buf, ae);
     accumulate_byte(buf, flags);
     accumulate_byte(buf, real_plen);
     accumulate_byte(buf, omit);
@@ -1856,7 +1898,7 @@ void
 buffer_wildcard_retraction(struct buffered *buf, struct interface *ifp)
 {
     start_message(buf, ifp, MESSAGE_UPDATE, 10);
-    accumulate_byte(buf, 0);
+    accumulate_byte(buf, AE_WILDCARD);
     accumulate_byte(buf, 0);
     accumulate_byte(buf, 0);
     accumulate_byte(buf, 0);
@@ -1941,7 +1983,7 @@ buffer_ihu(struct buffered *buf, struct interface *ifp, unsigned short rxcost,
     msglen = (ll ? 14 : 22) + (rtt_data ? 10 : 0);
 
     start_message(buf, ifp, MESSAGE_IHU, msglen);
-    accumulate_byte(buf, ll ? 3 : 2);
+    accumulate_byte(buf, ll ? AE_IPV6_LOCAL : AE_IPV6);
     accumulate_byte(buf, 0);
     accumulate_short(buf, rxcost);
     accumulate_short(buf, interval);
@@ -2061,7 +2103,7 @@ send_request(struct buffered *buf, struct interface *ifp,
         assert(!src_prefix);
         debugf("sending request for any.\n");
         start_message(buf, ifp, MESSAGE_REQUEST, 2);
-        accumulate_byte(buf, 0);
+        accumulate_byte(buf, AE_WILDCARD);
         accumulate_byte(buf, 0);
         end_message(buf, MESSAGE_REQUEST, 2);
         return;
@@ -2077,7 +2119,7 @@ send_request(struct buffered *buf, struct interface *ifp,
     len = 2 + pb + (is_ss ? 3 + spb : 0);
 
     start_message(buf, ifp, MESSAGE_REQUEST, len);
-    accumulate_byte(buf, v4 ? 1 : 2);
+    accumulate_byte(buf, v4 ? AE_IPV4 : AE_IPV6);
     accumulate_byte(buf, v4 ? plen - 96 : plen);
     if(v4)
         accumulate_bytes(buf, prefix + 12, pb);
@@ -2164,7 +2206,7 @@ send_multihop_request(struct buffered *buf, struct interface *ifp,
     len = 6 + 8 + pb + (is_ss ? 3 + spb : 0);
 
     start_message(buf, ifp, MESSAGE_MH_REQUEST, len);
-    accumulate_byte(buf, v4 ? 1 : 2);
+    accumulate_byte(buf, v4 ? AE_IPV4 : AE_IPV6);
     accumulate_byte(buf, v4 ? plen - 96 : plen);
     accumulate_short(buf, seqno);
     accumulate_byte(buf, hop_count);
